
/*-------------------------------------------------------------------*
 * Function decim_12k8() and oversamp_12k8()                         *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~                         *
 * decim_12k8    : decimation from fs to 12.8kHz.                    *
 * oversamp_12k8 : oversampling from 12.8kHz to fs.                  *
 *                                                                   *
 * fs = 16/22/24/28.8/32/44/48 kHz.                                  *
 *-------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "amr_plus_fx.h"
#include "typedef.h"
#include "basic_op.h"
#include "count.h"

#define L_FRAME_MAX   (L_FRAME24k)
#define L_FILT_MAX    (L_FILT24k)

#define L_SUBFR16k   80       /* Subframe size at 16kHz           */


void Decim_12k8_p(
  Word16 sig_fs[],     /* (i)  : signal to decimate      */
  Word16 lg,           /* (i)  : length of input         */
  Word16 sig12k8[],    /* (o)  : decimated signal        */
  Word16 mem[],        /* (i/o): memory (2*L_FILT_FS)    */
  Word16 band)         /* (i)  : 0=0..6.4k, 1=6.4..10.8k */
{
  Word16 i, ncoef;
  Word16 signal[(2*L_FILT_MAX)+L_FRAME_MAX];
  const Word16 *filter;
  Word16 gain;
  Word16 fac_up, fac_down, fac_down_up, ncoef2;

  gain = 1;  move16();
  test();test();test();
  switch (lg) {

  case L_FRAME16kPLUS:

    ncoef = ((NCOEF_32k+1)>>1);      move16();
    fac_up = FAC1_32k*2;      move16(); 
    fac_down = FAC2_32k;    move16();
    fac_down_up = 1;    move16();          /*5/(2*2)*/
    test();
    if (band == 0) 
    {
      filter = Filter_32k;          move16();
    }
    else 
    {
      filter = Filter_32k_7k;  move16();
    }    
    break;

  case L_FRAME24k:
    ncoef = ((NCOEF_48k+1)>>1);    move16();
    fac_up = FAC1_48k*2;    move16(); 
    fac_down = FAC2_48k;  move16();
    fac_down_up =  1;     move16();         /*15/ (4*2) */
    test();
    if (band == 0) 
    {
      filter = Filter_48k;          move16();
    }
    else 
    {
      filter = Filter_48k_hf;       move16();
    }

    break;

  case L_FRAME8k: /* This mode actually upsamples rather than down */
    ncoef = (((NCOEF_32k>>2)+1)>>1);    move16();
    fac_up = FAC1_32k*4;         move16();
    fac_down = FAC2_32k;         move16();
    fac_down_up = 0;              move16();   /*5/ (2*2*2)*/
    filter = Filter_8k;       move16();
    gain = add(gain,1);                 move16();                                       
    break;

  default:
    printf("wrong frame size in decim_12k8!\n");
    exit(0);
  }      

  ncoef2 = shl(ncoef,1);
  for (i=0; i<ncoef2; i++)
  {
    signal[i] = mem[i];         move16();
  }
  
  for (i=0; i<lg; i++) 
  {
    signal[i + ncoef2] = sig_fs[i];    move16();
  }

  Interpol(signal+ncoef, sig12k8, L_FRAME_PLUS, filter, ncoef, 
           fac_up,fac_down_up, fac_down, gain, 0);

  for (i=0; i<ncoef2; i++) 
  {
    mem[i] = signal[i+lg];        move16();
  }

  return;
}

