#include "typedef.h"
#include "basic_op.h"
#include "count.h"


#include "amr_plus_fx.h"
#include "stdlib.h"


/*-----------------------------------------------------------------*
 * Funtion  init_decoder_lf                                        *
 * ~~~~~~~~~~~~~~~~~~~~~~~~                                        *
 *                                                                 *
 *   ->Initialization of variables for the decoder section.        *
 *                                                                 *
 *-----------------------------------------------------------------*/

/* isp tables for initialization */

void Init_decoder_lf(Decoder_State_Plus_fx *st)
{
  Word32   i;
  /* Static vectors to zero */
  Set_zero(st->wold_exc, PIT_MAX_MAX+L_INTERPOL);

  Set_zero(st->wold_synth, M);
  Set_zero(st->cp_old_synth, M);


  Set_zero(st->wpast_isfq, M);
  Set_zero(st->wwovlp, L_OVLP);

  /* Initialize the ISFs */
  Copy(Isf_init, st->wisfold, M);

  for (i = 0; i < L_MEANBUF; i++)
  {
    Copy(Isf_init, &(st->wisf_buf[i*M]), M);
  }
  Copy(Isp_init, st->wispold, M);
       

  /* scaling memories for excitation */
  st->mem_subfr_q[6] = Q_MAX;  move16();
  st->mem_subfr_q[5] = Q_MAX;  move16();
  st->mem_subfr_q[4] = Q_MAX;  move16();
  st->mem_subfr_q[3] = Q_MAX;  move16();
  st->mem_subfr_q[2] = Q_MAX;  move16();
  st->mem_subfr_q[1] = Q_MAX;  move16();
  st->mem_subfr_q[0] = Q_MAX;  move16();
  
  Set_zero(st->old_subfr_q, 4*NB_DIV);
  Set_zero(st->old_syn_q, 4*NB_DIV);
  
  /* Initialize the ISPs */
  st->wold_T0 = 64;             move16();
  st->wold_T0_frac = 0;         move16();
  st->prev_lpc_lost = 0;        move16();
  st->seed_ace = 0;             move16();
  st->seed_tcx = 21845;         move16();
  st->pitch_tcx = L_DIV;        move16(); 

  st->wmem_wsyn = 0;            move16();
  st->wwsyn_rms = 0;            move16();
  st->wpast_gpit = 0;           move16();
  st->Lpast_gcode = 0L;         move32();

  st->Q_exc = 0;                move16();
  st->Q_syn = -1;               move16();
  st->Old_Q_syn = -1;           move16();


  st->prev_Q_syn = 6;           move16();
  st->Old_Q_exc = Q_MAX;        move16();
  st->Old_Qxri = 6;             move16();
  st->Old_Qxnq = Q_MAX;         move16();
  st->last_mode = 0;            move16();       /* set last mode to acelp */
  st->Old_QxnqMax = Q_MAX;      move16();
    
  st->ovlp_size = 0;            move16();
  st->wold_T0 = 64;             move16();
  st->wold_T0_frac = 0;         move16();
  st->i_offset = 0;             move16();

  Set_zero(st->mem_syn2,M);
  
  return;
}



/*-----------------------------------------------------------------*
 * Funtion decoder_lf                                              *
 * ~~~~~~~~~~~~~~~~~~                                              *
 *   ->Principle decoder routine (band 0 to 6400Hz).               *
 *                                                                 *
 *-----------------------------------------------------------------*/
void Decoder_lf(
  Word16 mod[],         /* (i)  : mode for each 20ms frame (mode[4]     */
  Word16 param[],       /* (i)  : parameters                            */
  Word16 nbits_AVQ[],   /* (i)  : for each frame (nbits_AVQ[4])         */
  Word16 codec_mode,    /* (i)  : AMR-WB+ mode (see cnst.h)             */ 
  Word16 bad_frame[],   /* (i)  : for each frame (bad_frame[4])         */
  Word16 AzLF[],        /* (o)  : decoded coefficients (AdLF[16])       */
  Word16 exc_[],        /* (o)  : decoded excitation                    */
  Word16 syn[],         /* (o)  : decoded synthesis                     */
  Word16 pitch[],       /* (o)  : decoded pitch (pitch[16])             */
  Word16 pit_gain[],    /* (o)  : decoded pitch gain (pit_gain[16])     */
  Word16 pit_adj,
  Decoder_State_Plus_fx *st /* (i/o) : coder memory state               */
)                       
{
  Word16 i, k, mode, bfi, bfi_2nd_st, *prm, tmp, tmp16;
  Word16 ispnew_[M], isfnew_[M], Az[(NB_SUBFR+1)*(M+1)];
  Word16 len;
  Word16 *synth_;
  Word16 stab_fac;
  Word32 Ltmp;

  test();
  if(pit_adj == 0) 
  {
    len = L_DIV;      move16();
  }
  else 
  {
    
   /*i = (((pit_adj*PIT_MIN_12k8)+(FSCALE_DENOM/2))/FSCALE_DENOM)-PIT_MIN_12k8;*/

    Ltmp = L_mac(FSCALE_DENOM, pit_adj, PIT_MIN_12k8);
  
    k = FSCALE_DENOM_NORMS;     move16();
 
    i = div_s(extract_h(L_shl(Ltmp, k)), shl(FSCALE_DENOM, k));
  
    tmp16 = add(i, 1);    
    Ltmp = L_msu(Ltmp, tmp16, FSCALE_DENOM);
  
    i = sub(i, PIT_MIN_12k8);
    test();
    if (Ltmp >= 0)
    {
      i = sub(tmp16,PIT_MIN_12k8);     
    } 

    st->i_offset = i; move16();

    len = add(shr(extract_l(L_mult(6,i)),1),PIT_MAX_12k8+L_INTERPOL);    move16();
    test();
    if (sub(len,L_DIV) < 0)
    {
      len = L_DIV;      move16();
    }
  }

  /* initialize bass postfilter */
  for (i=0; i<NB_SUBFR; i++) 
  {
    pitch[i] = L_SUBFR;         move16();   
    pit_gain[i] = 0;            move16();   
  }

  /* Initialize pointers */
  synth_ = syn;         move16();

  Copy(st->wold_synth, syn-M, M); 


  k = 0;            move16();
  test();
  move16(); /* initialize the pointer prm = param + (k*DEC_NPRM_DIV);  */
  move16(); /* initialize the counter k*(4*(M+1))  */
  move16(); /* initialize the counter k*L_DIV  */
  move16(); /* initialize the counter k*4  */

  while (sub(k,NB_DIV) < 0)
  {
    test();  
    mode = mod[k];              move16();

    bfi = bad_frame[k];         move16(); 

    bfi_2nd_st = 0;             move16();
    test();test();
    if ((sub(mode,2) == 0) && (bad_frame[k+1] != 0)) 
    {
      bfi_2nd_st = (1+2+4+8+16);        move16();
    }
    test(); 
    if (sub(mode,3) == 0) 
    {    
      test();
      if (bad_frame[k+1] != 0) 
      {
        bfi_2nd_st = 1;          move16();
      }
      test();
      if (bad_frame[k+2] != 0) 
      {
        bfi_2nd_st = add(bfi_2nd_st,(2+8));
      }
      test();
      if (bad_frame[k+3] != 0) 
      {
        bfi_2nd_st = add(bfi_2nd_st,(4+16));
      }
    }
    /* set pointer to parameters */
    prm = param + (k*DEC_NPRM_DIV); 


    /* decode ISFs and convert ISFs to cosine domain */
    Dpisf_2s_46b_2(prm, isfnew_,st->wpast_isfq, st->wisfold, st->wisf_buf, bfi, bfi_2nd_st, 1);

    test();test();
    if(bfi == 0 && bfi_2nd_st == 0)
    {
        st->seed_tcx = extract_l(L_shl(L_mac(L_mult(prm[0],prm[1]), prm[4],prm[5]),7));  move16();
    }
    
    prm += 7;

    Isf_isp(isfnew_, ispnew_, M);

    /* Check stability on isf : distance between old isf and current isf */

    Ltmp = 0;                             move32();
    for (i = 0; i < M - 1; i++)
    {
      tmp = sub(isfnew_[i], st->wisfold[i]);
      Ltmp = L_mac(Ltmp, tmp, tmp);
    }
    tmp = extract_h(L_shl(Ltmp, 8));
    tmp = mult(tmp, 26214);                /* tmp = Ltmp*0.8/256 */

    tmp = sub(20480, tmp);                 /* 1.25 - tmp */
    stab_fac = shl(tmp, 1);                /* Q14 -> Q15 with saturation */
    test();
    if (stab_fac < 0)
    {
      stab_fac = 0;                      move16();
    }

    /* ajust old isp[] following to a bad frame (to avoid overshoot) */
    test();test();
    if ((st->prev_lpc_lost != 0) && (bfi == 0)) 
    {
      for (i=0; i<M-1; i++) 
      {
        test();
        if (sub(ispnew_[i],st->wispold[i]) < 0) 
        {
          st->wispold[i] = ispnew_[i];          move16();  
        }
      }
    }
    /* - interpolate Ai in ISP domain (Aq) and save values for upper-band (Aq_lpc)
       - decode other parameters according to mode
       - set overlap size for next decoded frame (ovlp_size)
       - set mode for upper-band decoder (mod[])*/

    test();test();test();test();
    switch (mode) 
    {
 
      case 0:
      case 1:

      Int_lpc(st->wispold, ispnew_, interpol_frac4, Az, 4, M);


      Copy(Az, &AzLF[(k*4)*(M+1)], 5*(M+1));
      test();  
      if (mode == 0) /* 20-ms ACELP */
      { 
        /* set global variables used in dec_ace() even for bfi=1*/
        Decoder_acelp(prm, L_DIV, codec_mode, bad_frame[k],&pitch[k*4], 
                      &pit_gain[k*4], Az, &exc_[k*L_DIV], &synth_[k*L_DIV],pit_adj, len, stab_fac,st);


        /* average integer pitch-lag for high band coding */
        st->last_mode = 0;        move16() ; 
        Copy(&exc_[k*L_DIV],&exc_[k*L_DIV-(PIT_MAX_MAX + L_INTERPOL)], L_DIV);	
      }
      else
      {
        /* 20+2.5-ms TCX */
        tmp16 = st->Q_exc;    move16();
        Decoder_tcx(prm, &nbits_AVQ[k], Az, L_DIV, bad_frame+k,
                    &exc_[k*L_DIV], &synth_[k*L_DIV], st);
        st->last_mode =1;             move16();

        Copy(&exc_[k*L_DIV],&exc_[k*L_DIV-(PIT_MAX_MAX + L_INTERPOL)], L_DIV);
        /* update scaling on PIT_MAX_MAX - PIT_MAX samples (need in predlt4) */
        Scale_sig(&exc_[k*L_DIV+L_DIV-(PIT_MAX_MAX + L_INTERPOL)], (PIT_MAX_MAX + L_INTERPOL)-L_DIV, sub(st->Q_exc, tmp16));  /* Need to have an uniforme scaling over PIT_MAX_MAX */
      }
      Updt_mem_q(st->old_syn_q, 1, st->Q_syn);  
      Updt_mem_q(st->old_subfr_q, 1, st->Q_exc);

      k++;  /*ptr*/
      break;

    case 2: /* 40+5-ms TCX */

      Int_lpc(st->wispold, ispnew_, interpol_frac8, Az, 8, M);
      Copy(Az, &AzLF[(k*4)*(M+1)], 9*(M+1));

      Decoder_tcx(prm, &nbits_AVQ[k], Az, 2*L_DIV, bad_frame+k,
                    &exc_[k*L_DIV], &synth_[k*L_DIV], st);
      st->last_mode =2;                 move16();
      /* Update last subfr scaling, take in account the frame length*/
      Updt_mem_q(st->old_subfr_q, 2, st->Q_exc);
      Updt_mem_q(st->old_syn_q, 2, st->Q_syn);
      Copy(&exc_[k*L_DIV],&exc_[k*L_DIV-(PIT_MAX_MAX + L_INTERPOL)], 2*L_DIV);

      k+=2;         /*ptr*/
      break;

    case 3: /* 80+10-ms TCX */

      Int_lpc(st->wispold, ispnew_, interpol_frac16, Az, 16, M);
      Copy(Az, &AzLF[(k*4)*(M+1)], 17*(M+1));

      Decoder_tcx(prm, &nbits_AVQ[k], Az, 4*L_DIV, bad_frame+k,
                    &exc_[k*L_DIV], &synth_[k*L_DIV], st);
      st->last_mode =3;                    move16();

      /* Update last subfr scaling, take in account the frame length*/
      Updt_mem_q(st->old_subfr_q, 4, st->Q_exc);
      Updt_mem_q(st->old_syn_q, 4, st->Q_syn);

      Copy(&exc_[k*L_DIV],&exc_[k*L_DIV-(PIT_MAX_MAX + L_INTERPOL)], 4*L_DIV);

      k+=4;       /*ptr*/
      break;

    default:
      printf("decoder error: mode > 3!\n");
      exit(0);
    }

 

    st->prev_lpc_lost = bfi;           move16();

    /* update ispold[] and isfold[] for the next frame */
    Copy(ispnew_, st->wispold, M);
    Copy(isfnew_, st->wisfold, M);
              /* end of while*/
  }

  /*------ update signal for next superframe :
   shift exc[], synth[] and synth_hf[] to the left by L_FRAME_PLUS -----*/

  Copy(&syn[L_FRAME_PLUS-M], st->wold_synth, M);

  /* output exc and synth */
  /* Use no memorie excitation */
  Copy(synth_, syn, L_FRAME_PLUS);

  return;
}



